# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import hashlib
import http.client
import typing as ty
from unittest import mock
import uuid

import fixtures
import webtest

from keystone.auth import core as auth_core
from keystone.common import authorization
from keystone.common import context as keystone_context
from keystone.common import provider_api
from keystone.common import tokenless_auth
import keystone.conf
from keystone import exception
from keystone.federation import constants as federation_constants
from keystone.server.flask.request_processing.middleware import auth_context
from keystone.tests import unit
from keystone.tests.unit import mapping_fixtures
from keystone.tests.unit import test_backend_sql

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class MiddlewareRequestTestBase(unit.TestCase):
    MIDDLEWARE_CLASS: ty.Any = None  # override this in subclasses

    def _application(self):
        """A base wsgi application that returns a simple response."""

        def app(environ, start_response):
            # WSGI requires the body of the response to be bytes
            body = uuid.uuid4().hex.encode('utf-8')
            resp_headers = [
                ('Content-Type', 'text/html; charset=utf8'),
                ('Content-Length', str(len(body))),
            ]
            start_response('200 OK', resp_headers)
            return [body]

        return app

    def _generate_app_response(
        self, app, headers=None, method='get', path='/', **kwargs
    ):
        """Given a wsgi application wrap it in webtest and call it."""
        return getattr(webtest.TestApp(app), method)(
            path, headers=headers or {}, **kwargs
        )

    def _middleware_failure(self, exc, *args, **kwargs):
        """Assert that an exception is being thrown from process_request."""
        # NOTE(jamielennox): This is a little ugly. We need to call the webtest
        # framework so that the correct RequestClass object is created for when
        # we call process_request. However because we go via webtest we only
        # see the response object and not the actual exception that is thrown
        # by process_request. To get around this we subclass process_request
        # with something that checks for the right type of exception being
        # thrown so we can test the middle of the request process.
        # TODO(jamielennox): Change these tests to test the value of the
        # response rather than the error that is raised.

        class _Failing(self.MIDDLEWARE_CLASS):
            _called = False

            def fill_context(i_self, *i_args, **i_kwargs):
                # i_ to distinguish it from and not clobber the outer vars
                e = self.assertRaises(
                    exc, super().fill_context, *i_args, **i_kwargs
                )
                i_self._called = True
                raise e

        # by default the returned status when an uncaught exception is raised
        # for validation or caught errors this will likely be 400
        kwargs.setdefault('status', http.client.INTERNAL_SERVER_ERROR)  # 500

        app = _Failing(self._application())
        resp = self._generate_app_response(app, *args, **kwargs)
        self.assertTrue(app._called)
        return resp

    def _do_middleware_response(self, *args, **kwargs):
        """Wrap a middleware around a sample application and call it."""
        app = self.MIDDLEWARE_CLASS(self._application())
        return self._generate_app_response(app, *args, **kwargs)

    def _do_middleware_request(self, *args, **kwargs):
        """The request object from a successful middleware call."""
        return self._do_middleware_response(*args, **kwargs).request


class AuthContextMiddlewareTest(
    test_backend_sql.SqlTests, MiddlewareRequestTestBase
):
    MIDDLEWARE_CLASS = auth_context.AuthContextMiddleware

    def setUp(self):
        super().setUp()
        self.client_issuer = uuid.uuid4().hex
        self.untrusted_client_issuer = uuid.uuid4().hex
        self.trusted_issuer = self.client_issuer
        self.config_fixture.config(
            group='tokenless_auth', trusted_issuer=[self.trusted_issuer]
        )

        # client_issuer is encoded because you can't hash
        # unicode objects with hashlib.
        # This idp_id is calculated based on sha256(self.client_issuer)
        hashed_idp = hashlib.sha256(self.client_issuer.encode('utf-8'))
        self.idp_id = hashed_idp.hexdigest()
        self._load_sample_data()

    def _load_sample_data(self):
        self.protocol_id = 'x509'

        # 1) Create a domain for the user.
        self.domain = unit.new_domain_ref()
        self.domain_id = self.domain['id']
        self.domain_name = self.domain['name']
        PROVIDERS.resource_api.create_domain(self.domain_id, self.domain)

        # 2) Create a project for the user.
        self.project = unit.new_project_ref(domain_id=self.domain_id)
        self.project_id = self.project['id']
        self.project_name = self.project['name']

        PROVIDERS.resource_api.create_project(self.project_id, self.project)

        # 3) Create a user in new domain.
        self.user = unit.new_user_ref(
            domain_id=self.domain_id, project_id=self.project_id
        )

        self.user = PROVIDERS.identity_api.create_user(self.user)

        # Add IDP
        self.idp = self._idp_ref(id=self.idp_id)
        PROVIDERS.federation_api.create_idp(self.idp['id'], self.idp)

        # Add a role
        self.role = unit.new_role_ref()
        self.role_id = self.role['id']
        self.role_name = self.role['name']
        PROVIDERS.role_api.create_role(self.role_id, self.role)

        # Add a group
        self.group = unit.new_group_ref(domain_id=self.domain_id)
        self.group = PROVIDERS.identity_api.create_group(self.group)

        # Assign a role to the user on a project
        PROVIDERS.assignment_api.add_role_to_user_and_project(
            user_id=self.user['id'],
            project_id=self.project_id,
            role_id=self.role_id,
        )

        # Assign a role to the group on a project
        PROVIDERS.assignment_api.create_grant(
            role_id=self.role_id,
            group_id=self.group['id'],
            project_id=self.project_id,
        )

    def _load_mapping_rules(self, rules):
        # Add a mapping
        self.mapping = self._mapping_ref(rules=rules)
        PROVIDERS.federation_api.create_mapping(
            self.mapping['id'], self.mapping
        )
        # Add protocols
        self.proto_x509 = self._proto_ref(mapping_id=self.mapping['id'])
        self.proto_x509['id'] = self.protocol_id
        PROVIDERS.federation_api.create_protocol(
            self.idp['id'], self.proto_x509['id'], self.proto_x509
        )

    def _idp_ref(self, id=None):
        idp = {
            'id': id or uuid.uuid4().hex,
            'enabled': True,
            'description': uuid.uuid4().hex,
        }
        return idp

    def _proto_ref(self, mapping_id=None):
        proto = {
            'id': uuid.uuid4().hex,
            'mapping_id': mapping_id or uuid.uuid4().hex,
        }
        return proto

    def _mapping_ref(self, rules=None):
        if rules is None:
            mapped_rules = {}
        else:
            mapped_rules = rules.get('rules', {})
        return {
            'id': uuid.uuid4().hex,
            'rules': mapped_rules,
            'schema_version': "1.0",
        }

    def _assert_tokenless_auth_context(self, context, ephemeral_user=False):
        self.assertIsNotNone(context)
        self.assertEqual(self.project_id, context['project_id'])
        self.assertIn(self.role_name, context['roles'])
        if ephemeral_user:
            self.assertEqual(self.group['id'], context['group_ids'][0])
            self.assertEqual(
                'ephemeral', context[federation_constants.PROTOCOL]
            )
            self.assertEqual(
                self.idp_id, context[federation_constants.IDENTITY_PROVIDER]
            )
        else:
            self.assertEqual(self.user['id'], context['user_id'])

    def _assert_tokenless_request_context(
        self, request_context, ephemeral_user=False
    ):
        self.assertIsNotNone(request_context)
        self.assertEqual(self.project_id, request_context.project_id)
        self.assertIn(self.role_name, request_context.roles)
        if not ephemeral_user:
            self.assertEqual(self.user['id'], request_context.user_id)

    def test_context_already_exists(self):
        stub_value = uuid.uuid4().hex
        env = {authorization.AUTH_CONTEXT_ENV: stub_value}
        req = self._do_middleware_request(extra_environ=env)
        self.assertEqual(
            stub_value, req.environ.get(authorization.AUTH_CONTEXT_ENV)
        )

    def test_not_applicable_to_token_request(self):
        req = self._do_middleware_request(path='/auth/tokens', method='post')
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self.assertIsNone(context)

    def test_no_tokenless_attributes_request(self):
        req = self._do_middleware_request()
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self.assertIsNone(context)

    def test_no_issuer_attribute_request(self):
        env = {}
        env['HTTP_X_PROJECT_ID'] = uuid.uuid4().hex
        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self.assertIsNone(context)

    def test_has_only_issuer_and_project_name_request(self):
        env = {}
        # SSL_CLIENT_I_DN is the attribute name that wsgi env
        # references to issuer of the client certificate.
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = uuid.uuid4().hex
        self._middleware_failure(
            exception.ValidationError, extra_environ=env, status=400
        )

    def test_has_only_issuer_and_project_domain_name_request(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = uuid.uuid4().hex
        self._middleware_failure(
            exception.ValidationError, extra_environ=env, status=400
        )

    def test_has_only_issuer_and_project_domain_id_request(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_DOMAIN_ID'] = uuid.uuid4().hex
        self._middleware_failure(
            exception.ValidationError, extra_environ=env, status=400
        )

    def test_missing_both_domain_and_project_request(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        self._middleware_failure(
            exception.ValidationError, extra_environ=env, status=400
        )

    def test_empty_trusted_issuer_list(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_ID'] = uuid.uuid4().hex

        self.config_fixture.config(group='tokenless_auth', trusted_issuer=[])

        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self.assertIsNone(context)

    def test_client_issuer_not_trusted(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.untrusted_client_issuer
        env['HTTP_X_PROJECT_ID'] = uuid.uuid4().hex
        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self.assertIsNone(context)

    def test_proj_scope_with_proj_id_and_proj_dom_id_success(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_ID'] = self.project_id
        env['HTTP_X_PROJECT_DOMAIN_ID'] = self.domain_id
        # SSL_CLIENT_USER_NAME and SSL_CLIENT_DOMAIN_NAME are the types
        # defined in the mapping that will map to the user name and
        # domain name
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        env['SSL_CLIENT_DOMAIN_NAME'] = self.domain_name

        self._load_mapping_rules(
            mapping_fixtures.MAPPING_WITH_USERNAME_AND_DOMAINNAME
        )

        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self._assert_tokenless_auth_context(context)
        request_context = req.environ.get(keystone_context.REQUEST_CONTEXT_ENV)
        self._assert_tokenless_request_context(request_context)

    def test_proj_scope_with_proj_id_only_success(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_ID'] = self.project_id
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        env['SSL_CLIENT_DOMAIN_NAME'] = self.domain_name

        self._load_mapping_rules(
            mapping_fixtures.MAPPING_WITH_USERNAME_AND_DOMAINNAME
        )

        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self._assert_tokenless_auth_context(context)
        request_context = req.environ.get(keystone_context.REQUEST_CONTEXT_ENV)
        self._assert_tokenless_request_context(request_context)

    def test_proj_scope_with_proj_name_and_proj_dom_id_success(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_ID'] = self.domain_id
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        env['SSL_CLIENT_DOMAIN_NAME'] = self.domain_name

        self._load_mapping_rules(
            mapping_fixtures.MAPPING_WITH_USERNAME_AND_DOMAINNAME
        )

        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self._assert_tokenless_auth_context(context)
        request_context = req.environ.get(keystone_context.REQUEST_CONTEXT_ENV)
        self._assert_tokenless_request_context(request_context)

    def test_proj_scope_with_proj_name_and_proj_dom_name_success(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = self.domain_name
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        env['SSL_CLIENT_DOMAIN_NAME'] = self.domain_name

        self._load_mapping_rules(
            mapping_fixtures.MAPPING_WITH_USERNAME_AND_DOMAINNAME
        )

        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self._assert_tokenless_auth_context(context)
        request_context = req.environ.get(keystone_context.REQUEST_CONTEXT_ENV)
        self._assert_tokenless_request_context(request_context)

    def test_proj_scope_with_proj_name_only_fail(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_id
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        env['SSL_CLIENT_DOMAIN_NAME'] = self.domain_name

        self._load_mapping_rules(
            mapping_fixtures.MAPPING_WITH_USERNAME_AND_DOMAINNAME
        )

        self._middleware_failure(
            exception.ValidationError, extra_environ=env, status=400
        )

    def test_mapping_with_userid_and_domainid_success(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = self.domain_name
        env['SSL_CLIENT_USER_ID'] = self.user['id']
        env['SSL_CLIENT_DOMAIN_ID'] = self.domain_id

        self._load_mapping_rules(
            mapping_fixtures.MAPPING_WITH_USERID_AND_DOMAINID
        )

        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self._assert_tokenless_auth_context(context)
        request_context = req.environ.get(keystone_context.REQUEST_CONTEXT_ENV)
        self._assert_tokenless_request_context(request_context)

    def test_mapping_with_userid_and_domainname_success(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = self.domain_name
        env['SSL_CLIENT_USER_ID'] = self.user['id']
        env['SSL_CLIENT_DOMAIN_NAME'] = self.domain_name

        self._load_mapping_rules(
            mapping_fixtures.MAPPING_WITH_USERID_AND_DOMAINNAME
        )

        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self._assert_tokenless_auth_context(context)
        request_context = req.environ.get(keystone_context.REQUEST_CONTEXT_ENV)
        self._assert_tokenless_request_context(request_context)

    def test_mapping_with_username_and_domainid_success(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = self.domain_name
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        env['SSL_CLIENT_DOMAIN_ID'] = self.domain_id

        self._load_mapping_rules(
            mapping_fixtures.MAPPING_WITH_USERNAME_AND_DOMAINID
        )

        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self._assert_tokenless_auth_context(context)
        request_context = req.environ.get(keystone_context.REQUEST_CONTEXT_ENV)
        self._assert_tokenless_request_context(request_context)

    def test_only_domain_name_fail(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_ID'] = self.project_id
        env['HTTP_X_PROJECT_DOMAIN_ID'] = self.domain_id
        env['SSL_CLIENT_DOMAIN_NAME'] = self.domain_name

        self._load_mapping_rules(mapping_fixtures.MAPPING_WITH_DOMAINNAME_ONLY)

        self._middleware_failure(
            exception.ValidationError, extra_environ=env, status=400
        )

    def test_only_domain_id_fail(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_ID'] = self.project_id
        env['HTTP_X_PROJECT_DOMAIN_ID'] = self.domain_id
        env['SSL_CLIENT_DOMAIN_ID'] = self.domain_id

        self._load_mapping_rules(mapping_fixtures.MAPPING_WITH_DOMAINID_ONLY)

        self._middleware_failure(
            exception.ValidationError, extra_environ=env, status=400
        )

    def test_missing_domain_data_fail(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_ID'] = self.project_id
        env['HTTP_X_PROJECT_DOMAIN_ID'] = self.domain_id
        env['SSL_CLIENT_USER_NAME'] = self.user['name']

        self._load_mapping_rules(mapping_fixtures.MAPPING_WITH_USERNAME_ONLY)

        self._middleware_failure(
            exception.ValidationError, extra_environ=env, status=400
        )

    def test_userid_success(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_ID'] = self.project_id
        env['HTTP_X_PROJECT_DOMAIN_ID'] = self.domain_id
        env['SSL_CLIENT_USER_ID'] = self.user['id']

        self._load_mapping_rules(mapping_fixtures.MAPPING_WITH_USERID_ONLY)
        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self._assert_tokenless_auth_context(context)
        request_context = req.environ.get(keystone_context.REQUEST_CONTEXT_ENV)
        self._assert_tokenless_request_context(request_context)

    def test_domain_disable_fail(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = self.domain_name
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        env['SSL_CLIENT_DOMAIN_ID'] = self.domain_id

        self.domain['enabled'] = False
        self.domain = PROVIDERS.resource_api.update_domain(
            self.domain['id'], self.domain
        )

        self._load_mapping_rules(
            mapping_fixtures.MAPPING_WITH_USERNAME_AND_DOMAINID
        )
        self._middleware_failure(
            exception.Unauthorized, extra_environ=env, status=401
        )

    def test_user_disable_fail(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = self.domain_name
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        env['SSL_CLIENT_DOMAIN_ID'] = self.domain_id

        self.user['enabled'] = False
        self.user = PROVIDERS.identity_api.update_user(
            self.user['id'], self.user
        )

        self._load_mapping_rules(
            mapping_fixtures.MAPPING_WITH_USERNAME_AND_DOMAINID
        )

        self._middleware_failure(AssertionError, extra_environ=env)

    def test_invalid_user_fail(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_ID'] = self.project_id
        env['HTTP_X_PROJECT_DOMAIN_ID'] = self.domain_id
        env['SSL_CLIENT_USER_NAME'] = uuid.uuid4().hex
        env['SSL_CLIENT_DOMAIN_NAME'] = self.domain_name

        self._load_mapping_rules(
            mapping_fixtures.MAPPING_WITH_USERNAME_AND_DOMAINNAME
        )

        self._middleware_failure(
            exception.UserNotFound, extra_environ=env, status=404
        )

    def test_ephemeral_success(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = self.domain_name
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        self.config_fixture.config(
            group='tokenless_auth', protocol='ephemeral'
        )
        self.protocol_id = 'ephemeral'
        mapping = copy.deepcopy(mapping_fixtures.MAPPING_FOR_EPHEMERAL_USER)
        mapping['rules'][0]['local'][0]['group']['id'] = self.group['id']
        self._load_mapping_rules(mapping)

        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self._assert_tokenless_auth_context(context, ephemeral_user=True)
        request_context = req.environ.get(keystone_context.REQUEST_CONTEXT_ENV)
        self._assert_tokenless_request_context(
            request_context, ephemeral_user=True
        )

    def test_ephemeral_and_group_domain_name_mapping_success(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = self.domain_name
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        self.config_fixture.config(
            group='tokenless_auth', protocol='ephemeral'
        )
        self.protocol_id = 'ephemeral'
        mapping = copy.deepcopy(
            mapping_fixtures.MAPPING_FOR_EPHEMERAL_USER_AND_GROUP_DOMAIN_NAME
        )
        mapping['rules'][0]['local'][0]['group']['name'] = self.group['name']
        mapping['rules'][0]['local'][0]['group']['domain']['name'] = (
            self.domain['name']
        )
        self._load_mapping_rules(mapping)

        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self._assert_tokenless_auth_context(context, ephemeral_user=True)

    def test_ephemeral_with_default_user_type_success(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = self.domain_name
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        self.config_fixture.config(
            group='tokenless_auth', protocol='ephemeral'
        )
        self.protocol_id = 'ephemeral'
        # this mapping does not have the user type defined
        # and it should defaults to 'ephemeral' which is
        # the expected type for the test case.
        mapping = copy.deepcopy(
            mapping_fixtures.MAPPING_FOR_DEFAULT_EPHEMERAL_USER
        )
        mapping['rules'][0]['local'][0]['group']['id'] = self.group['id']
        self._load_mapping_rules(mapping)

        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self._assert_tokenless_auth_context(context, ephemeral_user=True)
        request_context = req.environ.get(keystone_context.REQUEST_CONTEXT_ENV)
        self._assert_tokenless_request_context(
            request_context, ephemeral_user=True
        )

    def test_ephemeral_any_user_success(self):
        """Verify ephemeral user does not need a specified user.

        Keystone is not looking to match the user, but a corresponding group.
        """
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = self.domain_name
        env['SSL_CLIENT_USER_NAME'] = uuid.uuid4().hex
        self.config_fixture.config(
            group='tokenless_auth', protocol='ephemeral'
        )
        self.protocol_id = 'ephemeral'
        mapping = copy.deepcopy(mapping_fixtures.MAPPING_FOR_EPHEMERAL_USER)
        mapping['rules'][0]['local'][0]['group']['id'] = self.group['id']
        self._load_mapping_rules(mapping)

        req = self._do_middleware_request(extra_environ=env)
        context = req.environ.get(authorization.AUTH_CONTEXT_ENV)
        self._assert_tokenless_auth_context(context, ephemeral_user=True)
        request_context = req.environ.get(keystone_context.REQUEST_CONTEXT_ENV)
        self._assert_tokenless_request_context(
            request_context, ephemeral_user=True
        )

    def test_ephemeral_invalid_scope_fail(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = uuid.uuid4().hex
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = uuid.uuid4().hex
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        self.config_fixture.config(
            group='tokenless_auth', protocol='ephemeral'
        )
        self.protocol_id = 'ephemeral'
        mapping = copy.deepcopy(mapping_fixtures.MAPPING_FOR_EPHEMERAL_USER)
        mapping['rules'][0]['local'][0]['group']['id'] = self.group['id']
        self._load_mapping_rules(mapping)

        self._middleware_failure(
            exception.Unauthorized, extra_environ=env, status=401
        )

    def test_ephemeral_no_group_found_fail(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = self.domain_name
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        self.config_fixture.config(
            group='tokenless_auth', protocol='ephemeral'
        )
        self.protocol_id = 'ephemeral'
        mapping = copy.deepcopy(mapping_fixtures.MAPPING_FOR_EPHEMERAL_USER)
        mapping['rules'][0]['local'][0]['group']['id'] = uuid.uuid4().hex
        self._load_mapping_rules(mapping)

        self._middleware_failure(
            exception.MappedGroupNotFound, extra_environ=env
        )

    def test_ephemeral_incorrect_mapping_fail(self):
        """Test ephemeral user picking up the non-ephemeral user mapping.

        Looking up the mapping with protocol Id 'x509' will load up
        the non-ephemeral user mapping, results unauthenticated.
        """
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        env['HTTP_X_PROJECT_NAME'] = self.project_name
        env['HTTP_X_PROJECT_DOMAIN_NAME'] = self.domain_name
        env['SSL_CLIENT_USER_NAME'] = self.user['name']
        # This will pick up the incorrect mapping
        self.config_fixture.config(group='tokenless_auth', protocol='x509')
        self.protocol_id = 'x509'
        mapping = copy.deepcopy(mapping_fixtures.MAPPING_FOR_EPHEMERAL_USER)
        mapping['rules'][0]['local'][0]['group']['id'] = uuid.uuid4().hex
        self._load_mapping_rules(mapping)

        self._middleware_failure(
            exception.MappedGroupNotFound, extra_environ=env
        )

    def test_create_idp_id_success(self):
        env = {}
        env['SSL_CLIENT_I_DN'] = self.client_issuer
        auth = tokenless_auth.TokenlessAuthHelper(env)
        idp_id = auth._build_idp_id()
        self.assertEqual(self.idp_id, idp_id)

    def test_create_idp_id_attri_not_found_fail(self):
        env = {}
        env[uuid.uuid4().hex] = self.client_issuer
        auth = tokenless_auth.TokenlessAuthHelper(env)
        expected_msg = (
            'Could not determine Identity Provider ID. The '
            f'configuration option {CONF.tokenless_auth.issuer_attribute} was not found in the '
            'request environment.'
        )
        # Check the content of the exception message as well
        self.assertRaisesRegex(
            exception.TokenlessAuthConfigError,
            expected_msg,
            auth._build_idp_id,
        )

    def test_admin_token_context(self):
        self.config_fixture.config(admin_token='ADMIN')
        log_fix = self.useFixture(fixtures.FakeLogger())
        headers = {authorization.AUTH_TOKEN_HEADER: 'ADMIN'}
        req = self._do_middleware_request(headers=headers)
        self.assertTrue(req.environ[auth_context.CONTEXT_ENV]['is_admin'])
        self.assertNotIn('Invalid user token', log_fix.output)

    def test_request_non_admin(self):
        self.config_fixture.config(admin_token='ADMIN')
        log_fix = self.useFixture(fixtures.FakeLogger())
        headers = {authorization.AUTH_TOKEN_HEADER: 'NOT-ADMIN'}
        self._do_middleware_request(headers=headers)
        self.assertIn('Invalid user token', log_fix.output)

    def test_token_is_cached(self):
        # Make sure we only call PROVIDERS.token_provider_api.validate_token()
        # once while in middleware so that we're mindful of performance
        context = auth_core.AuthContext(
            user_id=self.user['id'], methods=['password']
        )
        token = PROVIDERS.token_provider_api.issue_token(
            context['user_id'],
            context['methods'],
            project_id=self.project_id,
            auth_context=context,
        )
        headers = {authorization.AUTH_TOKEN_HEADER: token.id.encode('utf-8')}
        with mock.patch.object(
            PROVIDERS.token_provider_api, 'validate_token', return_value=token
        ) as token_mock:
            self._do_middleware_request(
                path='/v3/projects', method='get', headers=headers
            )
            token_mock.assert_called_once()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

key_repository = cfg.StrOpt(
    'key_repository',
    default='/etc/keystone/fernet-keys/',
    help=utils.fmt(
        """
Directory containing Fernet token keys. This directory must exist before using
`keystone-manage fernet_setup` for the first time, must be writable by the user
running `keystone-manage fernet_setup` or `keystone-manage fernet_rotate`, and
of course must be readable by keystone's server process. The repository may
contain keys in one of three states: a single staged key (always index 0) used
for token validation, a single primary key (always the highest index) used for
token creation and validation, and any number of secondary keys (all other
index values) used for token validation. With multiple keystone nodes, each
node must share the same key repository contents, with the exception of the
staged key (index 0). It is safe to run `keystone-manage fernet_rotate` once on
any one node to promote a staged key (index 0) to be the new primary
(incremented from the previous highest index), and produce a new staged key (a
new key with index 0); the resulting repository can then be atomically
replicated to other nodes without any risk of race conditions (for example, it
is safe to run `keystone-manage fernet_rotate` on host A, wait any amount of
time, create a tarball of the directory on host A, unpack it on host B to a
temporary location, and atomically move (`mv`) the directory into place on host
B). Running `keystone-manage fernet_rotate` *twice* on a key repository without
syncing other nodes will result in tokens that can not be validated by all
nodes.
"""
    ),
)

max_active_keys = cfg.IntOpt(
    'max_active_keys',
    default=3,
    min=1,
    help=utils.fmt(
        """
This controls how many keys are held in rotation by `keystone-manage
fernet_rotate` before they are discarded. The default value of 3 means that
keystone will maintain one staged key (always index 0), one primary key (the
highest numerical index), and one secondary key (every other index). Increasing
this value means that additional secondary keys will be kept in the rotation.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [key_repository, max_active_keys]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt(
        """
Entry point for the domain-specific configuration driver in the
`keystone.resource.domain_config` namespace. Only a `sql` option is provided by
keystone, so there is no reason to set this unless you are providing a custom
entry point.
"""
    ),
)

caching = cfg.BoolOpt(
    'caching',
    default=True,
    help=utils.fmt(
        """
Toggle for caching of the domain-specific configuration backend. This has no
effect unless global caching is enabled. There is normally no reason to disable
this.
"""
    ),
)

cache_time = cfg.IntOpt(
    'cache_time',
    default=300,
    help=utils.fmt(
        """
Time-to-live (TTL, in seconds) to cache domain-specific configuration data.
This has no effect unless `[domain_config] caching` is enabled.
"""
    ),
)

additional_whitelisted_options = cfg.Opt(
    'additional_whitelisted_options',
    type=cfg.types.Dict(value_type=cfg.types.List(bounds=True)),
    help=utils.fmt(
        """
Additional whitelisted domain-specific options for out-of-tree drivers.
This is a dictonary of lists with the key being the group name and value a list
of group options."""
    ),
)

additional_sensitive_options = cfg.Opt(
    'additional_sensitive_options',
    type=cfg.types.Dict(value_type=cfg.types.List(bounds=True)),
    help=utils.fmt(
        """
Additional sensitive domain-specific options for out-of-tree drivers.
This is a dictonary of lists with the key being the group name and value a list
of group options."""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [
    driver,
    caching,
    cache_time,
    additional_whitelisted_options,
    additional_sensitive_options,
]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Fix incorrect constraints.

Revision ID: 99de3849d860
Revises: e25ffa003242
Create Date: 2022-08-02 12:23:25.525035
"""

from alembic import op

# revision identifiers, used by Alembic.
revision = '99de3849d860'
down_revision = 'e25ffa003242'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('access_rule', schema=None) as batch_op:
        batch_op.drop_constraint('access_rule_external_id_key', type_='unique')

    with op.batch_alter_table('trust', schema=None) as batch_op:
        batch_op.drop_constraint(
            'duplicate_trust_constraint_expanded', type_='unique'
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

DEPRECATED_REASON = (
    "The service provider API is now aware of system scope and default roles."
)

deprecated_get_sp = policy.DeprecatedRule(
    name=base.IDENTITY % 'get_service_provider',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_list_sp = policy.DeprecatedRule(
    name=base.IDENTITY % 'list_service_providers',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_update_sp = policy.DeprecatedRule(
    name=base.IDENTITY % 'update_service_provider',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_create_sp = policy.DeprecatedRule(
    name=base.IDENTITY % 'create_service_provider',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)
deprecated_delete_sp = policy.DeprecatedRule(
    name=base.IDENTITY % 'delete_service_provider',
    check_str=base.RULE_ADMIN_REQUIRED,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.STEIN,
)


service_provider_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'create_service_provider',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Create federated service provider.',
        operations=[
            {
                'path': (
                    '/v3/OS-FEDERATION/service_providers/'
                    '{service_provider_id}'
                ),
                'method': 'PUT',
            }
        ],
        deprecated_rule=deprecated_create_sp,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'list_service_providers',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='List federated service providers.',
        operations=[
            {'path': '/v3/OS-FEDERATION/service_providers', 'method': 'GET'},
            {'path': '/v3/OS-FEDERATION/service_providers', 'method': 'HEAD'},
        ],
        deprecated_rule=deprecated_list_sp,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'get_service_provider',
        check_str=base.RULE_ADMIN_OR_SYSTEM_READER,
        scope_types=['system', 'project'],
        description='Get federated service provider.',
        operations=[
            {
                'path': (
                    '/v3/OS-FEDERATION/service_providers/'
                    '{service_provider_id}'
                ),
                'method': 'GET',
            },
            {
                'path': (
                    '/v3/OS-FEDERATION/service_providers/'
                    '{service_provider_id}'
                ),
                'method': 'HEAD',
            },
        ],
        deprecated_rule=deprecated_get_sp,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'update_service_provider',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Update federated service provider.',
        operations=[
            {
                'path': (
                    '/v3/OS-FEDERATION/service_providers/'
                    '{service_provider_id}'
                ),
                'method': 'PATCH',
            }
        ],
        deprecated_rule=deprecated_update_sp,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'delete_service_provider',
        check_str=base.RULE_ADMIN_REQUIRED,
        scope_types=['system', 'project'],
        description='Delete federated service provider.',
        operations=[
            {
                'path': (
                    '/v3/OS-FEDERATION/service_providers/'
                    '{service_provider_id}'
                ),
                'method': 'DELETE',
            }
        ],
        deprecated_rule=deprecated_delete_sp,
    ),
]


def list_rules():
    return service_provider_policies



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.common import sql
from keystone.endpoint_policy.backends import sql as sql_driver
from keystone.tests import unit
from keystone.tests.unit.backend import core_sql
from keystone.tests.unit.endpoint_policy.backends import test_base
from keystone.tests.unit.ksfixtures import database


class SQLModelTestCase(core_sql.BaseBackendSqlModels):
    """Test cases to validate the table structure."""

    def test_policy_association_model(self):
        cols = (
            ('id', sql.String, 64),
            ('policy_id', sql.String, 64),
            ('endpoint_id', sql.String, 64),
            ('service_id', sql.String, 64),
            ('region_id', sql.String, 64),
        )

        self.assertExpectedSchema('policy_association', cols)


class SQLDriverTestCase(test_base.DriverTestCase, unit.TestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(database.Database())
        self._driver = sql_driver.EndpointPolicy()

    @property
    def driver(self):
        return self._driver



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

import fixtures

from keystone.common import jwt_utils
from keystone.common import utils


class JWSKeyRepository(fixtures.Fixture):
    def __init__(self, config_fixture):
        super().__init__()
        self.config_fixture = config_fixture
        self.key_group = 'jwt_tokens'

    def setUp(self):
        super().setUp()

        # grab a couple of temporary directory file paths
        private_key_directory = self.useFixture(fixtures.TempDir()).path
        public_key_directory = self.useFixture(fixtures.TempDir()).path

        # set config to use temporary paths
        self.config_fixture.config(
            group=self.key_group,
            jws_private_key_repository=private_key_directory,
        )
        self.config_fixture.config(
            group=self.key_group,
            jws_public_key_repository=public_key_directory,
        )

        # create temporary repositories
        utils.create_directory(private_key_directory)
        utils.create_directory(public_key_directory)

        # create an asymmetric key pair for token signing and validation
        private_key_path = os.path.join(private_key_directory, 'private.pem')
        public_key_path = os.path.join(public_key_directory, 'public.pem')
        jwt_utils.create_jws_keypair(private_key_path, public_key_path)



# Copyright 2014 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystone.common import sql
from keystone.tests.unit import test_backend_endpoint_policy
from keystone.tests.unit import test_backend_sql


class SqlPolicyAssociationTable(test_backend_sql.SqlModels):
    """Set of tests for checking SQL Policy Association Mapping."""

    def test_policy_association_mapping(self):
        cols = (
            ('id', sql.String, 64),
            ('policy_id', sql.String, 64),
            ('endpoint_id', sql.String, 64),
            ('service_id', sql.String, 64),
            ('region_id', sql.String, 64),
        )
        self.assertExpectedSchema('policy_association', cols)


class SqlPolicyAssociationTests(
    test_backend_sql.SqlTests,
    test_backend_endpoint_policy.PolicyAssociationTests,
):
    def load_fixtures(self, fixtures):
        super().load_fixtures(fixtures)
        self.load_sample_data()



# Copyright 2014 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log

from keystone.common import manager
from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone.i18n import _

CONF = keystone.conf.CONF
LOG = log.getLogger(__name__)
PROVIDERS = provider_api.ProviderAPIs


class Manager(manager.Manager):
    """Default pivot point for the Endpoint Policy backend.

    See :mod:`keystone.common.manager.Manager` for more details on how this
    dynamically calls the backend.

    """

    driver_namespace = 'keystone.endpoint_policy'
    _provides_api = 'endpoint_policy_api'

    def __init__(self):
        super().__init__(CONF.endpoint_policy.driver)

    def _assert_valid_association(self, endpoint_id, service_id, region_id):
        """Assert that the association is supported.

        There are three types of association supported:

        - Endpoint (in which case service and region must be None)
        - Service and region (in which endpoint must be None)
        - Service (in which case endpoint and region must be None)

        """
        if (
            endpoint_id is not None
            and service_id is None
            and region_id is None
        ):
            return
        if (
            service_id is not None
            and region_id is not None
            and endpoint_id is None
        ):
            return
        if (
            service_id is not None
            and endpoint_id is None
            and region_id is None
        ):
            return

        raise exception.InvalidPolicyAssociation(
            endpoint_id=endpoint_id, service_id=service_id, region_id=region_id
        )

    def create_policy_association(
        self, policy_id, endpoint_id=None, service_id=None, region_id=None
    ):
        self._assert_valid_association(endpoint_id, service_id, region_id)
        self.driver.create_policy_association(
            policy_id, endpoint_id, service_id, region_id
        )

    def check_policy_association(
        self, policy_id, endpoint_id=None, service_id=None, region_id=None
    ):
        self._assert_valid_association(endpoint_id, service_id, region_id)
        self.driver.check_policy_association(
            policy_id, endpoint_id, service_id, region_id
        )

    def delete_policy_association(
        self, policy_id, endpoint_id=None, service_id=None, region_id=None
    ):
        self._assert_valid_association(endpoint_id, service_id, region_id)
        self.driver.delete_policy_association(
            policy_id, endpoint_id, service_id, region_id
        )

    def list_endpoints_for_policy(self, policy_id):
        def _get_endpoint(endpoint_id, policy_id):
            try:
                return PROVIDERS.catalog_api.get_endpoint(endpoint_id)
            except exception.EndpointNotFound:
                msg = (
                    'Endpoint %(endpoint_id)s referenced in '
                    'association for policy %(policy_id)s not found.'
                )
                LOG.warning(
                    msg, {'policy_id': policy_id, 'endpoint_id': endpoint_id}
                )
                raise

        def _get_endpoints_for_service(service_id, endpoints):
            # TODO(henry-nash): Consider optimizing this in the future by
            # adding an explicit list_endpoints_for_service to the catalog API.
            return [ep for ep in endpoints if ep['service_id'] == service_id]

        def _get_endpoints_for_service_and_region(
            service_id, region_id, endpoints, regions
        ):
            # TODO(henry-nash): Consider optimizing this in the future.
            # The lack of a two-way pointer in the region tree structure
            # makes this somewhat inefficient.

            def _recursively_get_endpoints_for_region(
                region_id,
                service_id,
                endpoint_list,
                region_list,
                endpoints_found,
                regions_examined,
            ):
                """Recursively search down a region tree for endpoints.

                :param region_id: the point in the tree to examine
                :param service_id: the service we are interested in
                :param endpoint_list: list of all endpoints
                :param region_list: list of all regions
                :param endpoints_found: list of matching endpoints found so
                                        far - which will be updated if more are
                                        found in this iteration
                :param regions_examined: list of regions we have already looked
                                         at - used to spot illegal circular
                                         references in the tree to avoid never
                                         completing search
                :returns: list of endpoints that match

                """
                if region_id in regions_examined:
                    msg = (
                        'Circular reference or a repeated entry found '
                        'in region tree - %(region_id)s.'
                    )
                    LOG.error(msg, {'region_id': ref.region_id})
                    return

                regions_examined.append(region_id)
                endpoints_found += [
                    ep
                    for ep in endpoint_list
                    if ep['service_id'] == service_id
                    and ep['region_id'] == region_id
                ]

                for region in region_list:
                    if region['parent_region_id'] == region_id:
                        _recursively_get_endpoints_for_region(
                            region['id'],
                            service_id,
                            endpoints,
                            regions,
                            endpoints_found,
                            regions_examined,
                        )

            endpoints_found = []
            regions_examined = []

            # Now walk down the region tree
            _recursively_get_endpoints_for_region(
                region_id,
                service_id,
                endpoints,
                regions,
                endpoints_found,
                regions_examined,
            )

            return endpoints_found

        matching_endpoints = []
        endpoints = PROVIDERS.catalog_api.list_endpoints()
        regions = PROVIDERS.catalog_api.list_regions()
        for ref in self.list_associations_for_policy(policy_id):
            if ref.get('endpoint_id') is not None:
                matching_endpoints.append(
                    _get_endpoint(ref['endpoint_id'], policy_id)
                )
                continue

            if (
                ref.get('service_id') is not None
                and ref.get('region_id') is None
            ):
                matching_endpoints += _get_endpoints_for_service(
                    ref['service_id'], endpoints
                )
                continue

            if (
                ref.get('service_id') is not None
                and ref.get('region_id') is not None
            ):
                matching_endpoints += _get_endpoints_for_service_and_region(
                    ref['service_id'], ref['region_id'], endpoints, regions
                )
                continue

            msg = (
                'Unsupported policy association found - '
                'Policy %(policy_id)s, Endpoint %(endpoint_id)s, '
                'Service %(service_id)s, Region %(region_id)s, '
            )
            LOG.warning(
                msg,
                {
                    'policy_id': policy_id,
                    'endpoint_id': ref['endpoint_id'],
                    'service_id': ref['service_id'],
                    'region_id': ref['region_id'],
                },
            )

        return matching_endpoints

    def get_policy_for_endpoint(self, endpoint_id):
        def _get_policy(policy_id, endpoint_id):
            try:
                return PROVIDERS.policy_api.get_policy(policy_id)
            except exception.PolicyNotFound:
                msg = (
                    'Policy %(policy_id)s referenced in association '
                    'for endpoint %(endpoint_id)s not found.'
                )
                LOG.warning(
                    msg, {'policy_id': policy_id, 'endpoint_id': endpoint_id}
                )
                raise

        def _look_for_policy_for_region_and_service(endpoint):
            """Look in the region and its parents for a policy.

            Examine the region of the endpoint for a policy appropriate for
            the service of the endpoint. If there isn't a match, then chase up
            the region tree to find one.

            """
            region_id = endpoint['region_id']
            regions_examined = []
            while region_id is not None:
                try:
                    ref = self.get_policy_association(
                        service_id=endpoint['service_id'], region_id=region_id
                    )
                    return ref['policy_id']
                except exception.PolicyAssociationNotFound:  # nosec
                    # There wasn't one for that region & service, handle below.
                    pass

                # There wasn't one for that region & service, let's
                # chase up the region tree
                regions_examined.append(region_id)
                region = PROVIDERS.catalog_api.get_region(region_id)
                region_id = None
                if region.get('parent_region_id') is not None:
                    region_id = region['parent_region_id']
                    if region_id in regions_examined:
                        msg = (
                            'Circular reference or a repeated entry '
                            'found in region tree - %(region_id)s.'
                        )
                        LOG.error(msg, {'region_id': region_id})
                        break

        # First let's see if there is a policy explicitly defined for
        # this endpoint.

        try:
            ref = self.get_policy_association(endpoint_id=endpoint_id)
            return _get_policy(ref['policy_id'], endpoint_id)
        except exception.PolicyAssociationNotFound:  # nosec
            # There wasn't a policy explicitly defined for this endpoint,
            # handled below.
            pass

        # There wasn't a policy explicitly defined for this endpoint, so
        # now let's see if there is one for the Region & Service.

        endpoint = PROVIDERS.catalog_api.get_endpoint(endpoint_id)
        policy_id = _look_for_policy_for_region_and_service(endpoint)
        if policy_id is not None:
            return _get_policy(policy_id, endpoint_id)

        # Finally, just check if there is one for the service.
        try:
            ref = self.get_policy_association(
                service_id=endpoint['service_id']
            )
            return _get_policy(ref['policy_id'], endpoint_id)
        except exception.PolicyAssociationNotFound:  # nosec
            # No policy is associated with endpoint, handled below.
            pass

        msg = _('No policy is associated with endpoint %(endpoint_id)s.') % {
            'endpoint_id': endpoint_id
        }
        raise exception.NotFound(msg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client
import uuid

from oslo_serialization import jsonutils

from keystone.common.policies import group as gp
from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import temporaryfile

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class _SystemUserGroupTests:
    """Common default functionality for all system users."""

    def test_user_can_list_groups(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            r = c.get('/v3/groups', headers=self.headers)
            self.assertEqual(1, len(r.json['groups']))
            self.assertEqual(group['id'], r.json['groups'][0]['id'])

    def test_user_can_get_a_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/groups/{}'.format(group['id']), headers=self.headers
            )
            self.assertEqual(group['id'], r.json['group']['id'])

    def test_user_can_list_group_members(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])

        with self.test_client() as c:
            r = c.get(
                '/v3/groups/{}/users'.format(group['id']), headers=self.headers
            )
            self.assertEqual(1, len(r.json['users']))
            self.assertEqual(user['id'], r.json['users'][0]['id'])

    def test_user_can_list_groups_for_other_users(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])

        with self.test_client() as c:
            r = c.get(
                '/v3/users/{}/groups'.format(user['id']), headers=self.headers
            )
            self.assertEqual(1, len(r.json['groups']))
            self.assertEqual(group['id'], r.json['groups'][0]['id'])

    def test_user_can_check_if_user_in_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])

        with self.test_client() as c:
            c.get(
                '/v3/groups/{}/users/{}'.format(group['id'], user['id']),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_user_cannot_get_non_existent_group_not_found(self):
        with self.test_client() as c:
            c.get(
                f'/v3/groups/{uuid.uuid4().hex}',
                headers=self.headers,
                expected_status_code=http.client.NOT_FOUND,
            )


class _SystemAndDomainMemberAndReaderGroupTests:
    """Common default functionality for system readers and system members."""

    def test_user_cannot_create_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        create = {
            'group': {'name': uuid.uuid4().hex, 'domain_id': domain['id']}
        }

        with self.test_client() as c:
            c.post(
                '/v3/groups',
                json=create,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )

        update = {'group': {'description': uuid.uuid4().hex}}

        with self.test_client() as c:
            c.patch(
                '/v3/groups/{}'.format(group['id']),
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            c.delete(
                '/v3/groups/{}'.format(group['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_add_users_to_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            c.put(
                '/v3/groups/{}/users/{}'.format(group['id'], user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_remove_users_from_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])

        with self.test_client() as c:
            c.delete(
                '/v3/groups/{}/users/{}'.format(group['id'], user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class SystemReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserGroupTests,
    _SystemAndDomainMemberAndReaderGroupTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_reader)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_reader['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserGroupTests,
    _SystemAndDomainMemberAndReaderGroupTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_member)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.member_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_member['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserGroupTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_create_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        create = {
            'group': {'name': uuid.uuid4().hex, 'domain_id': domain['id']}
        }

        with self.test_client() as c:
            c.post('/v3/groups', json=create, headers=self.headers)

    def test_user_can_update_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )

        update = {'group': {'description': uuid.uuid4().hex}}

        with self.test_client() as c:
            c.patch(
                '/v3/groups/{}'.format(group['id']),
                json=update,
                headers=self.headers,
            )

    def test_user_can_delete_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            c.delete('/v3/groups/{}'.format(group['id']), headers=self.headers)

    def test_user_can_add_users_to_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            c.put(
                '/v3/groups/{}/users/{}'.format(group['id'], user['id']),
                headers=self.headers,
            )

    def test_user_can_remove_users_from_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])

        with self.test_client() as c:
            c.delete(
                '/v3/groups/{}/users/{}'.format(group['id'], user['id']),
                headers=self.headers,
            )


class _DomainUserGroupTests:
    def test_user_can_list_groups_in_domain(self):
        # second domain
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        # one group in new domain
        group1 = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        # one group in user's domain
        group2 = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        # user should only see one group
        with self.test_client() as c:
            r = c.get('/v3/groups', headers=self.headers)
            self.assertEqual(1, len(r.json['groups']))
            self.assertNotIn(group1['id'], [g['id'] for g in r.json['groups']])
            self.assertEqual(group2['id'], r.json['groups'][0]['id'])

    def test_user_cannot_list_groups_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        with self.test_client() as c:
            r = c.get(
                '/v3/groups?domain_id={}'.format(domain['id']),
                headers=self.headers,
            )
            self.assertEqual(0, len(r.json['groups']))

    def test_user_can_get_group_in_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        with self.test_client() as c:
            r = c.get(
                '/v3/groups/{}'.format(group['id']), headers=self.headers
            )
            self.assertEqual(group['id'], r.json['group']['id'])

    def test_user_cannot_get_group_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        with self.test_client() as c:
            c.get(
                '/v3/groups/{}'.format(group['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_get_non_existent_group_forbidden(self):
        with self.test_client() as c:
            c.get(
                f'/v3/groups/{uuid.uuid4().hex}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_can_list_groups_in_domain_for_user_in_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])
        with self.test_client() as c:
            r = c.get(
                '/v3/users/{}/groups'.format(user['id']), headers=self.headers
            )
            self.assertEqual(1, len(r.json['groups']))
            self.assertEqual(group['id'], r.json['groups'][0]['id'])

    def test_user_cannot_list_groups_in_own_domain_user_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])
        with self.test_client() as c:
            c.get(
                '/v3/users/{}/groups'.format(user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_list_groups_for_non_existent_user_forbidden(self):
        with self.test_client() as c:
            c.get(
                f'/v3/users/{uuid.uuid4().hex}/groups',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_list_groups_in_other_domain_user_in_own_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        # one group in other domain
        group1 = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        # one group in own domain
        group2 = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        PROVIDERS.identity_api.add_user_to_group(user['id'], group1['id'])
        PROVIDERS.identity_api.add_user_to_group(user['id'], group2['id'])
        with self.test_client() as c:
            r = c.get(
                '/v3/users/{}/groups'.format(user['id']), headers=self.headers
            )
            # only one group should be visible
            self.assertEqual(1, len(r.json['groups']))
            self.assertEqual(group2['id'], r.json['groups'][0]['id'])

    def test_user_can_list_users_in_own_domain_for_group_in_own_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])
        with self.test_client() as c:
            r = c.get(
                '/v3/groups/{}/users'.format(group['id']), headers=self.headers
            )
            self.assertEqual(1, len(r.json['users']))
            self.assertEqual(user['id'], r.json['users'][0]['id'])

    def test_user_cannot_list_users_in_other_domain_group_in_own_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        # one user in other domain
        user1 = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )
        # one user in own domain
        user2 = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        PROVIDERS.identity_api.add_user_to_group(user1['id'], group['id'])
        PROVIDERS.identity_api.add_user_to_group(user2['id'], group['id'])
        with self.test_client() as c:
            r = c.get(
                '/v3/groups/{}/users'.format(group['id']), headers=self.headers
            )
            # only one user should be visible
            self.assertEqual(1, len(r.json['users']))
            self.assertEqual(user2['id'], r.json['users'][0]['id'])

    def test_user_cannot_list_users_in_own_domain_group_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])
        with self.test_client() as c:
            c.get(
                '/v3/groups/{}/users'.format(group['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_list_users_in_non_existent_group_forbidden(self):
        with self.test_client() as c:
            c.get(
                f'/v3/groups/{uuid.uuid4().hex}/users',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_can_check_user_in_own_domain_group_in_own_domain(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])
        with self.test_client() as c:
            c.head(
                '/v3/groups/{group}/users/{user}'.format(
                    group=group['id'], user=user['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )
            c.get(
                '/v3/groups/{group}/users/{user}'.format(
                    group=group['id'], user=user['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_user_cannot_check_user_in_other_domain_group_in_own_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])
        with self.test_client() as c:
            c.head(
                '/v3/groups/{group}/users/{user}'.format(
                    group=group['id'], user=user['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )
            c.get(
                '/v3/groups/{group}/users/{user}'.format(
                    group=group['id'], user=user['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class DomainReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainUserGroupTests,
    _SystemAndDomainMemberAndReaderGroupTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_admin = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def _override_policy(self):
        # TODO(cmurphy): Remove this once the deprecated policies in
        # keystone.common.policies.group have been removed. This is only
        # here to make sure we test the new policies instead of the deprecated
        # ones. Oslo.policy will OR deprecated policies with new policies to
        # maintain compatibility and give operators a chance to update
        # permissions or update policies without breaking users. This will
        # cause these specific tests to fail since we're trying to correct this
        # broken behavior with better scope checking.
        with open(self.policy_file_name, 'w') as f:
            overridden_policies = {
                'identity:get_group': gp.SYSTEM_READER_OR_DOMAIN_READER,
                'identity:list_groups': gp.SYSTEM_READER_OR_DOMAIN_READER,
                'identity:list_groups_for_user': gp.SYSTEM_READER_OR_DOMAIN_READER_FOR_TARGET_USER_OR_OWNER,
                'identity:list_users_in_group': gp.SYSTEM_READER_OR_DOMAIN_READER,
                'identity:check_user_in_group': gp.SYSTEM_READER_OR_DOMAIN_READER_FOR_TARGET_GROUP_USER,
            }
            f.write(jsonutils.dumps(overridden_policies))


class DomainMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainUserGroupTests,
    _SystemAndDomainMemberAndReaderGroupTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_admin = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def _override_policy(self):
        # TODO(cmurphy): Remove this once the deprecated policies in
        # keystone.common.policies.group have been removed. This is only
        # here to make sure we test the new policies instead of the deprecated
        # ones. Oslo.policy will OR deprecated policies with new policies to
        # maintain compatibility and give operators a chance to update
        # permissions or update policies without breaking users. This will
        # cause these specific tests to fail since we're trying to correct this
        # broken behavior with better scope checking.
        with open(self.policy_file_name, 'w') as f:
            overridden_policies = {
                'identity:get_group': gp.SYSTEM_READER_OR_DOMAIN_READER,
                'identity:list_groups': gp.SYSTEM_READER_OR_DOMAIN_READER,
                'identity:list_groups_for_user': gp.SYSTEM_READER_OR_DOMAIN_READER_FOR_TARGET_USER_OR_OWNER,
                'identity:list_users_in_group': gp.SYSTEM_READER_OR_DOMAIN_READER,
                'identity:check_user_in_group': gp.SYSTEM_READER_OR_DOMAIN_READER_FOR_TARGET_GROUP_USER,
            }
            f.write(jsonutils.dumps(overridden_policies))


class DomainAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainUserGroupTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_admin = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.admin_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def _override_policy(self):
        # TODO(cmurphy): Remove this once the deprecated policies in
        # keystone.common.policies.group have been removed. This is only
        # here to make sure we test the new policies instead of the deprecated
        # ones. Oslo.policy will OR deprecated policies with new policies to
        # maintain compatibility and give operators a chance to update
        # permissions or update policies without breaking users. This will
        # cause these specific tests to fail since we're trying to correct this
        # broken behavior with better scope checking.
        with open(self.policy_file_name, 'w') as f:
            overridden_policies = {
                'identity:get_group': gp.SYSTEM_READER_OR_DOMAIN_READER,
                'identity:list_groups': gp.SYSTEM_READER_OR_DOMAIN_READER,
                'identity:list_groups_for_user': gp.SYSTEM_READER_OR_DOMAIN_READER_FOR_TARGET_USER_OR_OWNER,
                'identity:create_group': gp.SYSTEM_ADMIN_OR_DOMAIN_ADMIN,
                'identity:update_group': gp.SYSTEM_ADMIN_OR_DOMAIN_ADMIN,
                'identity:delete_group': gp.SYSTEM_ADMIN_OR_DOMAIN_ADMIN,
                'identity:list_users_in_group': gp.SYSTEM_READER_OR_DOMAIN_READER,
                'identity:remove_user_from_group': gp.SYSTEM_ADMIN_OR_DOMAIN_ADMIN_FOR_TARGET_GROUP_USER,
                'identity:check_user_in_group': gp.SYSTEM_READER_OR_DOMAIN_READER_FOR_TARGET_GROUP_USER,
                'identity:add_user_to_group': gp.SYSTEM_ADMIN_OR_DOMAIN_ADMIN_FOR_TARGET_GROUP_USER,
            }
            f.write(jsonutils.dumps(overridden_policies))

    def test_user_can_create_group_for_own_domain(self):
        create = {
            'group': {'name': uuid.uuid4().hex, 'domain_id': self.domain_id}
        }

        with self.test_client() as c:
            c.post('/v3/groups', json=create, headers=self.headers)

    def test_user_cannot_create_group_for_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        create = {
            'group': {'name': uuid.uuid4().hex, 'domain_id': domain['id']}
        }

        with self.test_client() as c:
            c.post(
                '/v3/groups',
                json=create,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_can_update_group_in_own_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        update = {'group': {'description': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                '/v3/groups/{}'.format(group['id']),
                json=update,
                headers=self.headers,
            )

    def test_user_cannot_update_group_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )

        update = {'group': {'description': uuid.uuid4().hex}}
        with self.test_client() as c:
            c.patch(
                '/v3/groups/{}'.format(group['id']),
                json=update,
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_can_delete_group_in_own_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        with self.test_client() as c:
            c.delete('/v3/groups/{}'.format(group['id']), headers=self.headers)

    def test_user_cannot_delete_group_in_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        with self.test_client() as c:
            c.delete(
                '/v3/groups/{}'.format(group['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_can_remove_user_in_own_domain_from_group_in_own_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])
        with self.test_client() as c:
            c.delete(
                '/v3/groups/{group}/users/{user}'.format(
                    group=group['id'], user=user['id']
                ),
                headers=self.headers,
            )

    def test_user_cannot_remove_user_other_domain_from_group_own_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )
        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])
        with self.test_client() as c:
            c.delete(
                '/v3/groups/{group}/users/{user}'.format(
                    group=group['id'], user=user['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_remove_user_own_domain_from_group_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])
        with self.test_client() as c:
            c.delete(
                '/v3/groups/{group}/users/{user}'.format(
                    group=group['id'], user=user['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_remove_non_existent_user_from_group_forbidden(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        with self.test_client() as c:
            c.delete(
                '/v3/groups/{group}/users/{user}'.format(
                    group=group['id'], user=uuid.uuid4().hex
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_remove_user_from_non_existent_group_forbidden(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        with self.test_client() as c:
            c.delete(
                '/v3/groups/{group}/users/{user}'.format(
                    group=uuid.uuid4().hex, user=user['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_can_add_user_in_own_domain_to_group_in_own_domain(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        with self.test_client() as c:
            c.put(
                '/v3/groups/{group}/users/{user}'.format(
                    group=group['id'], user=user['id']
                ),
                headers=self.headers,
            )

    def test_user_cannot_add_user_other_domain_to_group_own_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )
        with self.test_client() as c:
            c.put(
                '/v3/groups/{group}/users/{user}'.format(
                    group=group['id'], user=user['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_add_user_own_domain_to_group_other_domain(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        with self.test_client() as c:
            c.put(
                '/v3/groups/{group}/users/{user}'.format(
                    group=group['id'], user=user['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_add_non_existent_user_to_group_forbidden(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )
        with self.test_client() as c:
            c.put(
                '/v3/groups/{group}/users/{user}'.format(
                    group=group['id'], user=uuid.uuid4().hex
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_add_user_from_non_existent_group_forbidden(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        with self.test_client() as c:
            c.put(
                '/v3/groups/{group}/users/{user}'.format(
                    group=uuid.uuid4().hex, user=user['id']
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class ProjectUserTests(
    base_classes.TestCaseWithBootstrap, common_auth.AuthTestMixin
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']

        user = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(user)['id']

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            project_id=project['id'],
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=user['password'],
            project_id=project['id'],
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_get_list_their_own_groups(self):
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=self.domain_id)
        )

        PROVIDERS.identity_api.add_user_to_group(self.user_id, group['id'])

        with self.test_client() as c:
            r = c.get(f'/v3/users/{self.user_id}/groups', headers=self.headers)
            self.assertEqual(1, len(r.json['groups']))
            self.assertEqual(group['id'], r.json['groups'][0]['id'])

    def test_user_cannot_list_groups_for_other_users(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])

        with self.test_client() as c:
            c.get(
                '/v3/users/{}/groups'.format(user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_list_groups(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            c.get(
                '/v3/groups',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_get_a_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )

        with self.test_client() as c:
            c.get(
                '/v3/groups/{}'.format(group['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_list_group_members(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])

        with self.test_client() as c:
            c.get(
                '/v3/groups/{}/users'.format(group['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_check_if_user_in_group(self):
        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=domain['id'])
        )
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=domain['id'])
        )

        PROVIDERS.identity_api.add_user_to_group(user['id'], group['id'])

        with self.test_client() as c:
            c.get(
                '/v3/groups/{}/users/{}'.format(group['id'], user['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_get_non_existent_group_forbidden(self):
        with self.test_client() as c:
            c.get(
                f'/v3/groups/{uuid.uuid4().hex}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import hashlib
import hmac
import http.client
import uuid

from keystone.api import s3tokens
from keystone.common import provider_api
from keystone import exception
from keystone.tests import unit
from keystone.tests.unit import test_v3

PROVIDERS = provider_api.ProviderAPIs


class S3ContribCore(test_v3.RestfulTestCase):
    def setUp(self):
        super().setUp()

        self.load_backends()

        self.cred_blob, self.credential = unit.new_ec2_credential(
            self.user['id'], self.project_id
        )
        PROVIDERS.credential_api.create_credential(
            self.credential['id'], self.credential
        )
        # Cert credential so that we can test when type != ec2
        self.cert_blob, self.cert_cred = unit.new_cert_credential(
            self.user['id'], self.project_id
        )
        PROVIDERS.credential_api.create_credential(
            self.cert_cred['id'], self.cert_cred
        )

    def test_http_get_method_not_allowed(self):
        resp = self.get(
            '/s3tokens',
            expected_status=http.client.METHOD_NOT_ALLOWED,
            convert=False,
        )
        self.assertEqual(http.client.METHOD_NOT_ALLOWED, resp.status_code)

    def _test_good_response(self, **kwargs):
        sts = 'string to sign'  # opaque string from swift3
        sig = hmac.new(
            self.cred_blob['secret'].encode('ascii'),
            sts.encode('ascii'),
            hashlib.sha1,
        ).digest()
        resp = self.post(
            '/s3tokens',
            body={
                'credentials': {
                    'access': self.cred_blob['access'],
                    'signature': base64.b64encode(sig).strip(),
                    'token': base64.b64encode(sts.encode('ascii')).strip(),
                }
            },
            expected_status=http.client.OK,
            **kwargs,
        )
        self.assertValidProjectScopedTokenResponse(
            resp, self.user, forbid_token_id=True
        )

    def test_good_response(self):
        self._test_good_response()

    def test_good_response_noauth(self):
        self._test_good_response(noauth=True)

    def test_bad_request(self):
        self.post(
            '/s3tokens', body={}, expected_status=http.client.BAD_REQUEST
        )

        self.post(
            '/s3tokens',
            body="not json",
            expected_status=http.client.BAD_REQUEST,
        )

        self.post('/s3tokens', expected_status=http.client.BAD_REQUEST)

    def test_bad_response(self):
        self.post(
            '/s3tokens',
            body={
                'credentials': {
                    'access': self.cred_blob['access'],
                    'signature': base64.b64encode(
                        b'totally not the sig'
                    ).strip(),
                    'token': base64.b64encode(b'string to sign').strip(),
                }
            },
            expected_status=http.client.UNAUTHORIZED,
        )

    def test_bad_credential_type(self):
        self.post(
            '/s3tokens',
            body={
                'credentials': {
                    'access': self.cert_blob['access'],
                    'signature': base64.b64encode(
                        b'totally not the sig'
                    ).strip(),
                    'token': base64.b64encode(b'string to sign').strip(),
                }
            },
            expected_status=http.client.NOT_FOUND,
        )

    def test_nonexistent_access_key(self):
        self.post(
            '/s3tokens',
            body={
                'credentials': {
                    'access': 'somelongaccesskeythatdoesnotexist',
                    'signature': base64.b64encode(
                        b'totally not the sig'
                    ).strip(),
                    'token': base64.b64encode(b'string to sign').strip(),
                }
            },
            expected_status=http.client.NOT_FOUND,
        )

    def test_good_signature_v1(self):
        creds_ref = {'secret': 'b121dd41cdcc42fe9f70e572e84295aa'}
        credentials = {
            'token': 'UFVUCjFCMk0yWThBc2dUcGdBbVk3UGhDZmc9PQphcHB'
            'saWNhdGlvbi9vY3RldC1zdHJlYW0KVHVlLCAxMSBEZWMgMjAxM'
            'iAyMTo0MTo0MSBHTVQKL2NvbnRfczMvdXBsb2FkZWRfZnJ'
            'vbV9zMy50eHQ=',
            'signature': 'IL4QLcLVaYgylF9iHj6Wb8BGZsw=',
        }

        self.assertIsNone(
            s3tokens.S3Resource._check_signature(creds_ref, credentials)
        )

    def test_bad_signature_v1(self):
        creds_ref = {'secret': 'b121dd41cdcc42fe9f70e572e84295aa'}
        credentials = {
            'token': 'UFVUCjFCMk0yWThBc2dUcGdBbVk3UGhDZmc9PQphcHB'
            'saWNhdGlvbi9vY3RldC1zdHJlYW0KVHVlLCAxMSBEZWMgMjAxM'
            'iAyMTo0MTo0MSBHTVQKL2NvbnRfczMvdXBsb2FkZWRfZnJ'
            'vbV9zMy50eHQ=',
            'signature': uuid.uuid4().hex,
        }

        self.assertRaises(
            exception.Unauthorized,
            s3tokens.S3Resource._check_signature,
            creds_ref,
            credentials,
        )

    def test_good_signature_v4(self):
        creds_ref = {'secret': 'e7a7a2240136494986991a6598d9fb9f'}
        credentials = {
            'token': 'QVdTNC1ITUFDLVNIQTI1NgoyMDE1MDgyNFQxMTIwNDFaCjIw'
            'MTUwODI0L1JlZ2lvbk9uZS9zMy9hd3M0X3JlcXVlc3QKZjIy'
            'MTU1ODBlZWI5YTE2NzM1MWJkOTNlODZjM2I2ZjA0YTkyOGY1'
            'YzU1MjBhMzkzNWE0NTM1NDBhMDk1NjRiNQ==',
            'signature': '730ba8f58df6ffeadd78f402e990b2910d60'
            'bc5c2aec63619734f096a4dd77be',
        }

        self.assertIsNone(
            s3tokens.S3Resource._check_signature(creds_ref, credentials)
        )

    def test_good_iam_signature_v4(self):
        creds_ref = {'secret': 'e7a7a2240136494986991a6598d9fb9f'}
        credentials = {
            'token': 'QVdTNC1ITUFDLVNIQTI1NgoyMDE1MDgyNFQxMTIwNDFaCjIw'
            'MTUwODI0L1JlZ2lvbk9uZS9pYW0vYXdzNF9yZXF1ZXN0CmYy'
            'MjE1NTgwZWViOWExNjczNTFiZDkzZTg2YzNiNmYwNGE5Mjhm'
            'NWM1NTIwYTM5MzVhNDUzNTQwYTA5NTY0YjU=',
            'signature': 'db4e15b3040f6afaa9d9d16002de2fc3425b'
            'eea0c6ea8c1b2bb674f052030b7d',
        }

        self.assertIsNone(
            s3tokens.S3Resource._check_signature(creds_ref, credentials)
        )

    def test_good_sts_signature_v4(self):
        creds_ref = {'secret': 'e7a7a2240136494986991a6598d9fb9f'}
        credentials = {
            'token': 'QVdTNC1ITUFDLVNIQTI1NgoyMDE1MDgyNFQxMTIwNDFaCjIw'
            'MTUwODI0L1JlZ2lvbk9uZS9zdHMvYXdzNF9yZXF1ZXN0CmYy'
            'MjE1NTgwZWViOWExNjczNTFiZDkzZTg2YzNiNmYwNGE5Mjhm'
            'NWM1NTIwYTM5MzVhNDUzNTQwYTA5NTY0YjU=',
            'signature': '3aa0b6f1414b92b2a32584068f83c6d09b7f'
            'daa11d4ea58912bbf1d8616ef56d',
        }

        self.assertIsNone(
            s3tokens.S3Resource._check_signature(creds_ref, credentials)
        )

    def test_bad_signature_v4(self):
        creds_ref = {'secret': 'e7a7a2240136494986991a6598d9fb9f'}
        # the signature is wrong on an otherwise correctly formed token
        credentials = {
            'token': 'QVdTNC1ITUFDLVNIQTI1NgoyMDE1MDgyNFQxMTIwNDFaCjIw'
            'MTUwODI0L1JlZ2lvbk9uZS9zMy9hd3M0X3JlcXVlc3QKZjIy'
            'MTU1ODBlZWI5YTE2NzM1MWJkOTNlODZjM2I2ZjA0YTkyOGY1'
            'YzU1MjBhMzkzNWE0NTM1NDBhMDk1NjRiNQ==',
            'signature': uuid.uuid4().hex,
        }

        self.assertRaises(
            exception.Unauthorized,
            s3tokens.S3Resource._check_signature,
            creds_ref,
            credentials,
        )

    def test_bad_service_v4(self):
        creds_ref = {'secret': 'e7a7a2240136494986991a6598d9fb9f'}
        # use 'bad' as the service scope instead of a recognised service
        credentials = {
            'token': 'QVdTNC1ITUFDLVNIQTI1NgoyMDE1MDgyNFQxMTIwNDFaCjIw'
            'MTUwODI0L1JlZ2lvbk9uZS9iYWQvYXdzNF9yZXF1ZXN0CmYy'
            'MjE1NTgwZWViOWExNjczNTFiZDkzZTg2YzNiNmYwNGE5Mjhm'
            'NWM1NTIwYTM5MzVhNDUzNTQwYTA5NTY0YjU=',
            'signature': '1a2dec50eb1bba97887d1103c2ead6a39911'
            '98c4be2537cf14d40b64cceb888b',
        }

        self.assertRaises(
            exception.Unauthorized,
            s3tokens.S3Resource._check_signature,
            creds_ref,
            credentials,
        )

    def test_bad_signing_key_v4(self):
        creds_ref = {'secret': 'e7a7a2240136494986991a6598d9fb9f'}
        # signed with aws4_badrequest instead of aws4_request
        credentials = {
            'token': 'QVdTNC1ITUFDLVNIQTI1NgoyMDE1MDgyNFQxMTIwNDFaCjIw'
            'MTUwODI0L1JlZ2lvbk9uZS9zMy9hd3M0X3JlcXVlc3QKZjIy'
            'MTU1ODBlZWI5YTE2NzM1MWJkOTNlODZjM2I2ZjA0YTkyOGY1'
            'YzU1MjBhMzkzNWE0NTM1NDBhMDk1NjRiNQ==',
            'signature': '52d02211a3767d00b2104ab28c9859003b0e'
            '9c8735cd10de7975f3b1212cca41',
        }

        self.assertRaises(
            exception.Unauthorized,
            s3tokens.S3Resource._check_signature,
            creds_ref,
            credentials,
        )

    def test_bad_short_scope_v4(self):
        creds_ref = {'secret': 'e7a7a2240136494986991a6598d9fb9f'}
        # credential scope has too few parts, missing final /aws4_request
        credentials = {
            'token': 'QVdTNC1ITUFDLVNIQTI1NgoyMDE1MDgyNFQxMTIwNDFaCjIw'
            'MTUwODI0L1JlZ2lvbk9uZS9zMwpmMjIxNTU4MGVlYjlhMTY3'
            'MzUxYmQ5M2U4NmMzYjZmMDRhOTI4ZjVjNTUyMGEzOTM1YTQ1'
            'MzU0MGEwOTU2NGI1',
            'signature': '28a075f1ee41e96c431153914998443ff0f5'
            '5fe93d31b37181f13ff4865942a2',
        }

        self.assertRaises(
            exception.Unauthorized,
            s3tokens.S3Resource._check_signature,
            creds_ref,
            credentials,
        )

    def test_bad_token_v4(self):
        creds_ref = {'secret': 'e7a7a2240136494986991a6598d9fb9f'}
        # token has invalid format of first part
        credentials = {'token': 'QVdTNC1BQUEKWApYClg=', 'signature': ''}
        self.assertRaises(
            exception.Unauthorized,
            s3tokens.S3Resource._check_signature,
            creds_ref,
            credentials,
        )

        # token has invalid format of scope
        credentials = {
            'token': 'QVdTNC1ITUFDLVNIQTI1NgpYCi8vczMvYXdzTl9yZXF1ZXN0Clg=',
            'signature': '',
        }
        self.assertRaises(
            exception.Unauthorized,
            s3tokens.S3Resource._check_signature,
            creds_ref,
            credentials,
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fixtures
from oslo_policy import opts

from keystone.common.rbac_enforcer import policy


class Policy(fixtures.Fixture):
    """A fixture for working with policy configuration."""

    def __init__(self, config_fixture, policy_file=None):
        self._policy_file = policy_file
        self._config_fixture = config_fixture

    def setUp(self):
        super().setUp()
        opts.set_defaults(self._config_fixture.conf)
        self._config_fixture.config(
            group='oslo_policy', policy_file=self._policy_file
        )
        policy._ENFORCER.suppress_deprecation_warnings = True
        self.addCleanup(policy.reset)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from keystone.conf import utils

driver = cfg.StrOpt(
    'driver',
    default='sql',
    help=utils.fmt(
        """
Entry point for the token revocation backend driver in the `keystone.revoke`
namespace. Keystone only provides a `sql` driver, so there is no reason to set
this option unless you are providing a custom entry point.
"""
    ),
)

expiration_buffer = cfg.IntOpt(
    'expiration_buffer',
    default=1800,
    min=0,
    help=utils.fmt(
        """
The number of seconds after a token has expired before a corresponding
revocation event may be purged from the backend.
"""
    ),
)

caching = cfg.BoolOpt(
    'caching',
    default=True,
    help=utils.fmt(
        """
Toggle for revocation event caching. This has no effect unless global caching
is enabled.
"""
    ),
)

cache_time = cfg.IntOpt(
    'cache_time',
    default=3600,
    deprecated_opts=[
        cfg.DeprecatedOpt('revocation_cache_time', group='token')
    ],
    help=utils.fmt(
        """
Time to cache the revocation list and the revocation events (in seconds). This
has no effect unless global and `[revoke] caching` are both enabled.
"""
    ),
)


GROUP_NAME = __name__.split('.')[-1]
ALL_OPTS = [driver, expiration_buffer, caching, cache_time]


def register_opts(conf):
    conf.register_opts(ALL_OPTS, group=GROUP_NAME)


def list_opts():
    return {GROUP_NAME: ALL_OPTS}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This file handles all flask-restful resources for /v3/OS-TRUST

# TODO(morgan): Deprecate /v3/OS-TRUST/trusts path in favour of /v3/trusts.
# /v3/OS-TRUST should remain indefinitely.

import http.client

import flask
import flask_restful
from oslo_log import log
from oslo_policy import _checks as op_checks

from keystone.api._shared import json_home_relations
from keystone.api import validation
from keystone.common import context
from keystone.common import json_home
from keystone.common import provider_api
from keystone.common import rbac_enforcer
from keystone.common.rbac_enforcer import policy
from keystone.common import utils
from keystone import exception
from keystone.i18n import _
from keystone.server import flask as ks_flask
from keystone.trust import schema

LOG = log.getLogger(__name__)
ENFORCER = rbac_enforcer.RBACEnforcer
PROVIDERS = provider_api.ProviderAPIs

_build_resource_relation = json_home_relations.os_trust_resource_rel_func
_build_parameter_relation = json_home_relations.os_trust_parameter_rel_func

TRUST_ID_PARAMETER_RELATION = _build_parameter_relation(
    parameter_name='trust_id'
)


def _build_trust_target_enforcement():
    target = {}
    # NOTE(cmurphy) unlike other APIs, in the event the trust doesn't exist or
    # has 0 remaining uses, we actually do expect it to return a 404 and not a
    # 403, so don't catch NotFound here (lp#1840288)
    target['trust'] = PROVIDERS.trust_api.get_trust(
        flask.request.view_args.get('trust_id')
    )

    return target


def _trustor_trustee_only(trust):
    user_id = flask.request.environ.get(context.REQUEST_CONTEXT_ENV).user_id
    if user_id not in [
        trust.get('trustee_user_id'),
        trust.get('trustor_user_id'),
    ]:
        raise exception.ForbiddenAction(
            action=_('Requested user has no relation to this trust')
        )


def _normalize_trust_expires_at(trust):
    # correct isotime
    if trust.get('expires_at') is not None:
        trust['expires_at'] = utils.isotime(
            trust['expires_at'], subsecond=True
        )


def _normalize_trust_roles(trust):
    # fill in role data
    trust_full_roles = []
    for trust_role in trust.get('roles', []):
        trust_role = trust_role['id']
        try:
            matching_role = PROVIDERS.role_api.get_role(trust_role)
            full_role = ks_flask.ResourceBase.wrap_member(
                matching_role, collection_name='roles', member_name='role'
            )
            trust_full_roles.append(full_role['role'])
        except exception.RoleNotFound:
            pass

    trust['roles'] = trust_full_roles
    trust['roles_links'] = {
        'self': ks_flask.base_url(path='/{}/roles'.format(trust['id'])),
        'next': None,
        'previous': None,
    }


class TrustResourceBase(ks_flask.ResourceBase):
    def _check_unrestricted(self):
        if self.oslo_context.is_admin:
            return
        token = self.auth_context['token']
        if 'application_credential' in token.methods:
            if not token.application_credential['unrestricted']:
                action = _(
                    "Using method 'application_credential' is not "
                    "allowed for managing trusts."
                )
                raise exception.ForbiddenAction(action=action)


class TrustsResource(TrustResourceBase):
    collection_key = 'trusts'
    member_key = 'trust'
    api_prefix = '/OS-TRUST'
    json_home_resource_rel_func = _build_resource_relation
    json_home_parameter_rel_func = _build_parameter_relation

    def _find_redelegated_trust(self):
        # Check if delegated via trust
        redelegated_trust = None
        if self.oslo_context.is_delegated_auth:
            src_trust_id = self.oslo_context.trust_id
            if not src_trust_id:
                action = _('Redelegation allowed for delegated by trust only')
                raise exception.ForbiddenAction(action=action)
            redelegated_trust = PROVIDERS.trust_api.get_trust(src_trust_id)
        return redelegated_trust

    @staticmethod
    def _parse_expiration_date(expiration_date):
        if expiration_date is not None:
            return utils.parse_expiration_date(expiration_date)
        return None

    def _require_trustor_has_role_in_project(self, trust):
        trustor_roles = self._get_trustor_roles(trust)
        for trust_role in trust['roles']:
            matching_roles = [
                x for x in trustor_roles if x == trust_role['id']
            ]
            if not matching_roles:
                raise exception.RoleNotFound(role_id=trust_role['id'])

    def _get_trustor_roles(self, trust):
        original_trust = trust.copy()
        while original_trust.get('redelegated_trust_id'):
            original_trust = PROVIDERS.trust_api.get_trust(
                original_trust['redelegated_trust_id']
            )

        if (trust.get('project_id')) not in [None, '']:
            # Check project exists.
            PROVIDERS.resource_api.get_project(trust['project_id'])
            # Get a list of roles including any domain specific roles
            assignment_list = PROVIDERS.assignment_api.list_role_assignments(
                user_id=original_trust['trustor_user_id'],
                project_id=original_trust['project_id'],
                effective=True,
                strip_domain_roles=False,
            )
            return list({x['role_id'] for x in assignment_list})
        else:
            return []

    def _normalize_role_list(self, trust_roles):
        roles = []
        for role in trust_roles:
            if role.get('id'):
                roles.append({'id': role['id']})
            else:
                roles.append(
                    PROVIDERS.role_api.get_unique_role_by_name(role['name'])
                )
        return roles

    @validation.request_query_schema(schema.trust_index_request_query)
    @validation.response_body_schema(schema.trust_index_response_body)
    def get(self):
        """Dispatch for LIST trusts.

        GET /v3/OS-TRUST/trusts
        """
        trustor_user_id = flask.request.args.get('trustor_user_id')
        trustee_user_id = flask.request.args.get('trustee_user_id')
        if trustor_user_id:
            target = {'trust': {'trustor_user_id': trustor_user_id}}
            ENFORCER.enforce_call(
                action='identity:list_trusts_for_trustor', target_attr=target
            )
        elif trustee_user_id:
            target = {'trust': {'trustee_user_id': trustee_user_id}}
            ENFORCER.enforce_call(
                action='identity:list_trusts_for_trustee', target_attr=target
            )
        else:
            ENFORCER.enforce_call(action='identity:list_trusts')

        trusts = []

        # NOTE(cmurphy) As of Train, the default policies enforce the
        # identity:list_trusts rule and there are new policies in-code to
        # enforce identity:list_trusts_for_trustor and
        # identity:list_trusts_for_trustee. However, in case the
        # identity:list_trusts rule has been locally overridden by the default
        # that would have been produced by the sample config, we need to
        # enforce it again and warn that the behavior is changing.
        rules = policy._ENFORCER._enforcer.rules.get('identity:list_trusts')
        # rule check_str is ""
        if isinstance(rules, op_checks.TrueCheck):
            LOG.warning(
                "The policy check string for rule \"identity:list_trusts\" "
                "has been overridden to \"always true\". In the next release, "
                "this will cause the \"identity:list_trusts\" action to be "
                "fully permissive as hardcoded enforcement will be removed. "
                "To correct this issue, either stop overriding the "
                "\"identity:list_trusts\" rule in config to accept the "
                "defaults, or explicitly set a rule that is not empty."
            )
            if not flask.request.args:
                # NOTE(morgan): Admin can list all trusts.
                ENFORCER.enforce_call(action='admin_required')

        if not flask.request.args:
            trusts += PROVIDERS.trust_api.list_trusts()
        elif trustor_user_id:
            trusts += PROVIDERS.trust_api.list_trusts_for_trustor(
                trustor_user_id
            )
        elif trustee_user_id:
            trusts += PROVIDERS.trust_api.list_trusts_for_trustee(
                trustee_user_id
            )

        for trust in trusts:
            # get_trust returns roles, list_trusts does not
            # It seems in some circumstances, roles does not
            # exist in the query response, so check first
            if 'roles' in trust:
                del trust['roles']

            if trust.get('expires_at') is not None:
                trust['expires_at'] = utils.isotime(
                    trust['expires_at'], subsecond=True
                )

        return self.wrap_collection(trusts)

    @validation.request_body_schema(schema.trust_create_request_body)
    @validation.response_body_schema(schema.trust_response_body)
    def post(self):
        """Create a new trust.

        The User creating the trust must be the trustor.

        POST /v3/OS-TRUST/trusts
        """
        ENFORCER.enforce_call(action='identity:create_trust')
        trust = self.request_body_json.get('trust', {})
        self._check_unrestricted()

        if trust.get('project_id') and not trust.get('roles'):
            action = _('At least one role should be specified')
            raise exception.ForbiddenAction(action=action)

        if self.oslo_context.user_id != trust.get('trustor_user_id'):
            action = _("The authenticated user should match the trustor")
            raise exception.ForbiddenAction(action=action)

        # Ensure the trustee exists
        PROVIDERS.identity_api.get_user(trust['trustee_user_id'])

        # Normalize roles
        trust['roles'] = self._normalize_role_list(trust.get('roles', []))
        self._require_trustor_has_role_in_project(trust)
        trust['expires_at'] = self._parse_expiration_date(
            trust.get('expires_at')
        )
        trust = self._assign_unique_id(trust)
        redelegated_trust = self._find_redelegated_trust()
        return_trust = PROVIDERS.trust_api.create_trust(
            trust_id=trust['id'],
            trust=trust,
            roles=trust['roles'],
            redelegated_trust=redelegated_trust,
            initiator=self.audit_initiator,
        )
        _normalize_trust_expires_at(return_trust)
        _normalize_trust_roles(return_trust)
        return self.wrap_member(return_trust), http.client.CREATED


class TrustResource(TrustResourceBase):
    collection_key = 'trusts'
    member_key = 'trust'
    api_prefix = '/OS-TRUST'
    json_home_resource_rel_func = _build_resource_relation
    json_home_parameter_rel_func = _build_parameter_relation

    @validation.request_query_schema(schema.trust_request_query)
    @validation.response_body_schema(schema.trust_response_body)
    def get(self, trust_id):
        """Get trust.

        GET /v3/OS-TRUST/trusts/{trust_id}
        """
        ENFORCER.enforce_call(
            action='identity:get_trust',
            build_target=_build_trust_target_enforcement,
        )

        # NOTE(cmurphy) look up trust before doing is_admin authorization - to
        # maintain the API contract, we expect a missing trust to raise a 404
        # before we get to enforcement (lp#1840288)
        trust = PROVIDERS.trust_api.get_trust(trust_id)

        if self.oslo_context.is_admin:
            # policies are not loaded for the is_admin context, so need to
            # block access here
            raise exception.ForbiddenAction(
                action=_('Requested user has no relation to this trust')
            )

        # NOTE(cmurphy) As of Train, the default policies enforce the
        # identity:get_trust rule. However, in case the
        # identity:get_trust rule has been locally overridden by the
        # default that would have been produced by the sample config, we need
        # to enforce it again and warn that the behavior is changing.
        rules = policy._ENFORCER._enforcer.rules.get('identity:get_trust')
        # rule check_str is ""
        if isinstance(rules, op_checks.TrueCheck):
            LOG.warning(
                "The policy check string for rule \"identity:get_trust\" "
                "has been overridden to \"always true\". In the next release, "
                "this will cause the"
                "\"identity:get_trust\" action to "
                "be fully permissive as hardcoded enforcement will be "
                "removed. To correct this issue, either stop overriding the "
                "\"identity:get_trust\" rule in config to accept the "
                "defaults, or explicitly set a rule that is not empty."
            )
            _trustor_trustee_only(trust)

        _normalize_trust_expires_at(trust)
        _normalize_trust_roles(trust)
        return self.wrap_member(trust)

    def delete(self, trust_id):
        """Delete trust.

        DELETE /v3/OS-TRUST/trusts/{trust_id}
        """
        ENFORCER.enforce_call(
            action='identity:delete_trust',
            build_target=_build_trust_target_enforcement,
        )
        self._check_unrestricted()

        # NOTE(cmurphy) As of Train, the default policies enforce the
        # identity:delete_trust rule. However, in case the
        # identity:delete_trust rule has been locally overridden by the
        # default that would have been produced by the sample config, we need
        # to enforce it again and warn that the behavior is changing.
        rules = policy._ENFORCER._enforcer.rules.get('identity:delete_trust')
        # rule check_str is ""
        if isinstance(rules, op_checks.TrueCheck):
            LOG.warning(
                "The policy check string for rule \"identity:delete_trust\" "
                "has been overridden to \"always true\". In the next release, "
                "this will cause the"
                "\"identity:delete_trust\" action to "
                "be fully permissive as hardcoded enforcement will be "
                "removed. To correct this issue, either stop overriding the "
                "\"identity:delete_trust\" rule in config to accept the "
                "defaults, or explicitly set a rule that is not empty."
            )
            trust = PROVIDERS.trust_api.get_trust(trust_id)
            if (
                self.oslo_context.user_id != trust.get('trustor_user_id')
                and not self.oslo_context.is_admin
            ):
                action = _('Only admin or trustor can delete a trust')
                raise exception.ForbiddenAction(action=action)
        PROVIDERS.trust_api.delete_trust(
            trust_id, initiator=self.audit_initiator
        )
        return '', http.client.NO_CONTENT


# NOTE(morgan): Since this Resource is not being used with the automatic
# URL additions and does not have a collection key/member_key, we use
# the flask-restful Resource, not the keystone ResourceBase
class RolesForTrustListResource(flask_restful.Resource):
    @property
    def oslo_context(self):
        return flask.request.environ.get(context.REQUEST_CONTEXT_ENV, None)

    def get(self, trust_id):
        ENFORCER.enforce_call(
            action='identity:list_roles_for_trust',
            build_target=_build_trust_target_enforcement,
        )

        # NOTE(morgan): This duplicates a little of the .get_trust from the
        # main resource, as it needs some of the same logic. However, due to
        # how flask-restful works, this should be fully encapsulated

        if self.oslo_context.is_admin:
            # policies are not loaded for the is_admin context, so need to
            # block access here
            raise exception.ForbiddenAction(
                action=_('Requested user has no relation to this trust')
            )

        trust = PROVIDERS.trust_api.get_trust(trust_id)

        # NOTE(cmurphy) As of Train, the default policies enforce the
        # identity:list_roles_for_trust rule. However, in case the
        # identity:list_roles_for_trust rule has been locally overridden by the
        # default that would have been produced by the sample config, we need
        # to enforce it again and warn that the behavior is changing.
        rules = policy._ENFORCER._enforcer.rules.get(
            'identity:list_roles_for_trust'
        )
        # rule check_str is ""
        if isinstance(rules, op_checks.TrueCheck):
            LOG.warning(
                "The policy check string for rule "
                "\"identity:list_roles_for_trust\" has been overridden to "
                "\"always true\". In the next release, this will cause the "
                "\"identity:list_roles_for_trust\" action to be fully "
                "permissive as hardcoded enforcement will be removed. To "
                "correct this issue, either stop overriding the "
                "\"identity:get_trust\" rule in config to accept the "
                "defaults, or explicitly set a rule that is not empty."
            )
            _trustor_trustee_only(trust)

        _normalize_trust_expires_at(trust)
        _normalize_trust_roles(trust)
        return {'roles': trust['roles'], 'links': trust['roles_links']}


# NOTE(morgan): Since this Resource is not being used with the automatic
# URL additions and does not have a collection key/member_key, we use
# the flask-restful Resource, not the keystone ResourceBase
class RoleForTrustResource(flask_restful.Resource):
    @property
    def oslo_context(self):
        return flask.request.environ.get(context.REQUEST_CONTEXT_ENV, None)

    def get(self, trust_id, role_id):
        """Get a role that has been assigned to a trust."""
        ENFORCER.enforce_call(
            action='identity:get_role_for_trust',
            build_target=_build_trust_target_enforcement,
        )

        if self.oslo_context.is_admin:
            # policies are not loaded for the is_admin context, so need to
            # block access here
            raise exception.ForbiddenAction(
                action=_('Requested user has no relation to this trust')
            )

        trust = PROVIDERS.trust_api.get_trust(trust_id)

        # NOTE(cmurphy) As of Train, the default policies enforce the
        # identity:get_role_for_trust rule. However, in case the
        # identity:get_role_for_trust rule has been locally overridden by the
        # default that would have been produced by the sample config, we need
        # to enforce it again and warn that the behavior is changing.
        rules = policy._ENFORCER._enforcer.rules.get(
            'identity:get_role_for_trust'
        )
        # rule check_str is ""
        if isinstance(rules, op_checks.TrueCheck):
            LOG.warning(
                "The policy check string for rule "
                "\"identity:get_role_for_trust\" has been overridden to "
                "\"always true\". In the next release, this will cause the "
                "\"identity:get_role_for_trust\" action to be fully "
                "permissive as hardcoded enforcement will be removed. To "
                "correct this issue, either stop overriding the "
                "\"identity:get_role_for_trust\" rule in config to accept the "
                "defaults, or explicitly set a rule that is not empty."
            )
            _trustor_trustee_only(trust)

        if not any(role['id'] == role_id for role in trust['roles']):
            raise exception.RoleNotFound(role_id=role_id)

        role = PROVIDERS.role_api.get_role(role_id)
        return ks_flask.ResourceBase.wrap_member(
            role, collection_name='roles', member_name='role'
        )


class TrustAPI(ks_flask.APIBase):
    _name = 'trusts'
    _import_name = __name__
    resource_mapping = [
        ks_flask.construct_resource_map(
            resource=TrustsResource,
            url='/trusts',
            resource_kwargs={},
            rel="trusts",
            path_vars=None,
            resource_relation_func=_build_resource_relation,
        ),
        ks_flask.construct_resource_map(
            resource=TrustResource,
            url='/trusts/<string:trust_id>',
            resource_kwargs={},
            rel="trust",
            path_vars={'trust_id': TRUST_ID_PARAMETER_RELATION},
            resource_relation_func=_build_resource_relation,
        ),
        ks_flask.construct_resource_map(
            resource=RolesForTrustListResource,
            url='/trusts/<string:trust_id>/roles',
            resource_kwargs={},
            rel='trust_roles',
            path_vars={'trust_id': TRUST_ID_PARAMETER_RELATION},
            resource_relation_func=_build_resource_relation,
        ),
        ks_flask.construct_resource_map(
            resource=RoleForTrustResource,
            url='/trusts/<string:trust_id>/roles/<string:role_id>',
            resource_kwargs={},
            rel='trust_role',
            path_vars={
                'trust_id': TRUST_ID_PARAMETER_RELATION,
                'role_id': json_home.Parameters.ROLE_ID,
            },
            resource_relation_func=_build_resource_relation,
        ),
    ]
    _api_url_prefix = '/OS-TRUST'


APIs = (TrustAPI,)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Useful utilities for tests."""

import functools
import os
import time
import unittest
import uuid

TZ = None


def timezone(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        tz_original = os.environ.get('TZ')
        try:
            if TZ:
                os.environ['TZ'] = TZ
                time.tzset()
            return func(*args, **kwargs)
        finally:
            if TZ:
                if tz_original:
                    os.environ['TZ'] = tz_original
                else:
                    if 'TZ' in os.environ:
                        del os.environ['TZ']
                time.tzset()

    return wrapper


def new_uuid():
    """Return a string UUID."""
    return uuid.uuid4().hex


def wip(message, expected_exception=Exception, bug=None):
    """Mark a test as work in progress.

    Based on code by Nat Pryce:
    https://gist.github.com/npryce/997195#file-wip-py

    The test will always be run. If the test fails then a SkipTest
    exception is raised. If the test passes an AssertionError exception
    is raised so that the developer knows they made the test pass. This
    is a reminder to remove the decorator.

    :param message: a string message to help clarify why the test is
                    marked as a work in progress
    :param expected_exception: an exception class that will be checked for
                               when @wip verifies an exception is raised. The
                               test will fail if a different exception is
                               raised. Default is "any" exception is valid
    :param bug: (optional) a string for tracking the bug and what bug should
                cause the @wip decorator to be removed from the testcase

    Usage:
      >>> @wip('Expected Error', expected_exception=Exception, bug="#000000")
      >>> def test():
      >>>     pass

    """
    if bug:
        bugstr = " (BugID " + bug + ")"
    else:
        bugstr = ""

    def _wip(f):
        @functools.wraps(f)
        def run_test(*args, **kwargs):
            __e = None
            try:
                f(*args, **kwargs)
            except Exception as __e:  # noqa F841
                if expected_exception != Exception and not isinstance(
                    __e, expected_exception
                ):
                    raise AssertionError(
                        f'Work In Progress Test Failed{bugstr} with '
                        f'unexpected exception. Expected "{expected_exception.__class__.__name__}" '
                        f'got "{__e.__class__.__name__}": {message} '
                    )
                # NOTE(notmorgan): We got the expected exception we can safely
                # skip this test.
                raise unittest.SkipTest(
                    'Work In Progress Test Failed as '
                    f'expected{bugstr}: {message}'
                )

            raise AssertionError(
                f'Work In Progress Test Passed{bugstr}: ' f'{message}'
            )

        return run_test

    return _wip



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystone import exception


class DriverTestCase:
    """Test cases to validate the current policy driver behavior."""

    def setUp(self):
        super().setUp()

        self.policy = {
            'id': uuid.uuid4().hex,
            'blob': '{"identity:create_user": "role:domain_admin"}',
            'type': 'application/json',
        }
        self.driver.create_policy(self.policy['id'], self.policy)

    @property
    def driver(self):
        raise exception.NotImplemented()

    def test_list_policies(self):
        another_policy = {
            'id': uuid.uuid4().hex,
            'blob': '{"compute:create": "role:project_member"}',
            'type': 'application/json',
        }
        self.driver.create_policy(another_policy['id'], another_policy)

        policies = self.driver.list_policies()

        self.assertCountEqual([self.policy, another_policy], policies)

    def test_get_policy(self):
        self.assertEqual(
            self.policy, self.driver.get_policy(self.policy['id'])
        )

    def test_update_policy(self):
        self.policy['blob'] = (
            '{"identity:create_user": "role:domain_admin",'
            '"identity:update_user": "role:domain_admin"}'
        )

        self.driver.update_policy(self.policy['id'], self.policy)

        self.assertEqual(
            self.policy, self.driver.get_policy(self.policy['id'])
        )

    def test_delete_policy(self):
        self.driver.delete_policy(self.policy['id'])

        self.assertRaises(
            exception.PolicyNotFound, self.driver.get_policy, self.policy['id']
        )
